"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Overlay = void 0;
const printTree_1 = require("tree-dump/lib/printTree");
const printBinary_1 = require("tree-dump/lib/printBinary");
const util_1 = require("sonic-forest/lib/util");
const util_2 = require("sonic-forest/lib/splay/util");
const OverlayPoint_1 = require("./OverlayPoint");
const MarkerOverlayPoint_1 = require("./MarkerOverlayPoint");
const refs_1 = require("./refs");
const clock_1 = require("../../../json-crdt-patch/clock");
const json_hash_1 = require("../../../json-hash");
const MarkerSlice_1 = require("../slice/MarkerSlice");
const iterator_1 = require("../../../util/iterator");
/**
 * Overlay is a tree structure that represents all the intersections of slices
 * in the text. It is used to quickly find all the slices that overlap a
 * given point in the text. The overlay is a read-only structure, its state
 * is changed only by calling the `refresh` method, which updates the overlay
 * based on the current state of the text and slices.
 */
class Overlay {
    constructor(txt) {
        this.txt = txt;
        this.root = undefined;
        this.leadingTextHash = 0;
        // ----------------------------------------------------------------- Stateful
        this.hash = 0;
        this.slices = new Map();
        this.start = this.point(this.txt.str.id, 1 /* Anchor.After */);
    }
    point(id, anchor) {
        return new OverlayPoint_1.OverlayPoint(this.txt.str, id, anchor);
    }
    mPoint(marker, anchor) {
        return new MarkerOverlayPoint_1.MarkerOverlayPoint(this.txt.str, marker.start.id, anchor, marker);
    }
    first() {
        return this.root ? (0, util_1.first)(this.root) : undefined;
    }
    last() {
        return this.root ? (0, util_1.last)(this.root) : undefined;
    }
    iterator() {
        let curr = this.first();
        return () => {
            const ret = curr;
            if (curr)
                curr = (0, util_1.next)(curr);
            return ret;
        };
    }
    entries() {
        return new iterator_1.UndefEndIter(this.iterator());
    }
    [Symbol.iterator]() {
        return this.entries();
    }
    markerIterator() {
        let curr = this.first();
        return () => {
            while (curr) {
                const ret = curr;
                if (curr)
                    curr = (0, util_1.next)(curr);
                if (ret instanceof MarkerOverlayPoint_1.MarkerOverlayPoint)
                    return ret;
            }
            return;
        };
    }
    markers() {
        return new iterator_1.UndefEndIter(this.iterator());
    }
    /**
     * Retrieve overlay point or the previous one, measured in spacial dimension.
     */
    getOrNextLower(point) {
        if (point.isAbsStart()) {
            const first = this.first();
            if (!first)
                return;
            if (first.isAbsStart())
                return first;
            point = first;
        }
        else if (point.isAbsEnd())
            return this.last();
        let curr = this.root;
        let result = undefined;
        while (curr) {
            const cmp = curr.cmpSpatial(point);
            if (cmp === 0)
                return curr;
            if (cmp > 0)
                curr = curr.l;
            else {
                const next = curr.r;
                result = curr;
                if (!next)
                    return result;
                curr = next;
            }
        }
        return result;
    }
    /**
     * Retrieve overlay point or the next one, measured in spacial dimension.
     */
    getOrNextHigher(point) {
        if (point.isAbsEnd()) {
            const last = this.last();
            if (!last)
                return;
            if (last.isAbsEnd())
                return last;
            point = last;
        }
        else if (point.isAbsStart())
            return this.first();
        let curr = this.root;
        let result = undefined;
        while (curr) {
            const cmp = curr.cmpSpatial(point);
            if (cmp === 0)
                return curr;
            if (cmp < 0)
                curr = curr.r;
            else {
                const next = curr.l;
                result = curr;
                if (!next)
                    return result;
                curr = next;
            }
        }
        return result;
    }
    find(predicate) {
        let point = this.first();
        while (point) {
            if (predicate(point))
                return point;
            point = (0, util_1.next)(point);
        }
        return;
    }
    chunkSlices0(chunk, p1, p2, callback) {
        const rga = this.txt.str;
        const strId = rga.id;
        let checkFirstAnchor = p1.anchor === 1 /* Anchor.After */;
        const adjustForLastAnchor = p2.anchor === 0 /* Anchor.Before */;
        let id1 = p1.id;
        const id1IsStr = !(0, clock_1.compare)(id1, strId);
        if (id1IsStr) {
            const first = rga.first();
            if (!first)
                return;
            id1 = first.id;
            checkFirstAnchor = false;
        }
        const id2 = p2.id;
        if (!checkFirstAnchor && !adjustForLastAnchor) {
            return rga.range0(chunk, id1, id2, callback);
        }
        const sid1 = id1.sid;
        const time1 = id1.time;
        const sid2 = id2.sid;
        const time2 = id2.time;
        return rga.range0(undefined, id1, id2, (chunk, off, len) => {
            if (checkFirstAnchor) {
                checkFirstAnchor = false;
                const chunkId = chunk.id;
                if (chunkId.sid === sid1 && chunkId.time + off === time1) {
                    if (len <= 1)
                        return;
                    off += 1;
                    len -= 1;
                }
            }
            if (adjustForLastAnchor) {
                const chunkId = chunk.id;
                if (chunkId.sid === sid2 && chunkId.time + off + len - 1 === time2) {
                    if (len <= 1)
                        return;
                    len -= 1;
                }
            }
            callback(chunk, off, len);
        });
    }
    points0(start, end, callback) {
        const txt = this.txt;
        const str = txt.str;
        const strFirstChunk = str.first();
        if (!strFirstChunk)
            return;
        let point = start || this.first();
        let prev;
        const pointIsStart = point &&
            ((!(0, clock_1.compare)(point.id, str.id) && point.anchor === 1 /* Anchor.After */) ||
                (!(0, clock_1.compare)(strFirstChunk.id, point.id) && point.anchor === 0 /* Anchor.Before */));
        if (!start && !pointIsStart) {
            const startPoint = this.start;
            startPoint.id = strFirstChunk.id;
            startPoint.anchor = 0 /* Anchor.Before */;
            callback(startPoint);
        }
        while (point) {
            if (end && end(point))
                return;
            callback(point);
            prev = point;
            point = (0, util_1.next)(point);
        }
        const strLastChunk = str.last();
        const strLastChunkId = strLastChunk.id;
        if (prev) {
            const prevId = prev.id;
            if (prev.anchor === 1 /* Anchor.After */ &&
                prevId.time === strLastChunkId.time + strLastChunk.span - 1 &&
                prevId.sid === strLastChunkId.sid &&
                prevId.sid === strLastChunkId.sid)
                return;
        }
        const endId = strLastChunk.span > 1 ? (0, clock_1.tick)(strLastChunkId, strLastChunk.span - 1) : strLastChunkId;
        const ending = this.point(endId, 1 /* Anchor.After */);
        if (end && end(ending))
            return;
        callback(ending);
    }
    points1(start, end, callback) {
        let p1;
        let p2;
        this.points0(start, end, (point) => {
            if (p1) {
                p2 = point;
                callback(p1, p2);
                p1 = p2;
            }
            else {
                p1 = point;
            }
        });
    }
    findContained(range) {
        const result = new Set();
        let point = this.getOrNextLower(range.start);
        if (!point)
            return result;
        do {
            if (!range.containsPoint(point))
                continue;
            const slices = point.layers;
            const length = slices.length;
            for (let i = 0; i < length; i++) {
                const slice = slices[i];
                if (!result.has(slice) && range.contains(slice))
                    result.add(slice);
            }
            if (point instanceof MarkerOverlayPoint_1.MarkerOverlayPoint) {
                const marker = point.marker;
                if (marker && !result.has(marker) && range.contains(marker))
                    result.add(marker);
            }
        } while (point && (point = (0, util_1.next)(point)) && range.containsPoint(point));
        return result;
    }
    findOverlapping(range) {
        const result = new Set();
        let point = this.getOrNextLower(range.start);
        if (!point)
            return result;
        do {
            const slices = point.layers;
            const length = slices.length;
            for (let i = 0; i < length; i++)
                result.add(slices[i]);
            if (point instanceof MarkerOverlayPoint_1.MarkerOverlayPoint) {
                const marker = point.marker;
                if (marker)
                    result.add(marker);
            }
        } while (point && (point = (0, util_1.next)(point)) && range.containsPoint(point));
        return result;
    }
    computeSplitTextHashes() {
        const txt = this.txt;
        const str = txt.str;
        const firstChunk = str.first();
        if (!firstChunk)
            return;
        let chunk = firstChunk;
        let marker = undefined;
        let state = 5381 /* CONST.START_STATE */;
        this.points1(undefined, undefined, (p1, p2) => {
            // TODO: need to incorporate slice attribute hash here?
            const id1 = p1.id;
            state = (state << 5) + state + (id1.sid >>> 0) + id1.time;
            let overlayPointHash = 5381 /* CONST.START_STATE */;
            chunk = this.chunkSlices0(chunk || firstChunk, p1, p2, (chunk, off, len) => {
                const id = chunk.id;
                overlayPointHash =
                    (overlayPointHash << 5) + overlayPointHash + ((((id.sid >>> 0) + id.time) << 8) + (off << 4) + len);
            });
            state = (0, json_hash_1.updateNum)(state, overlayPointHash);
            if (p1) {
                p1.hash = overlayPointHash;
            }
            if (p2 instanceof MarkerOverlayPoint_1.MarkerOverlayPoint) {
                if (marker) {
                    marker.textHash = state;
                }
                else {
                    this.leadingTextHash = state;
                }
                state = 5381 /* CONST.START_STATE */;
                marker = p2;
            }
        });
        if (marker instanceof MarkerOverlayPoint_1.MarkerOverlayPoint) {
            marker.textHash = state;
        }
        else {
            this.leadingTextHash = state;
        }
    }
    isBlockSplit(id) {
        const point = this.txt.point(id, 0 /* Anchor.Before */);
        const overlayPoint = this.getOrNextLower(point);
        return (overlayPoint instanceof MarkerOverlayPoint_1.MarkerOverlayPoint && overlayPoint.id.time === id.time && overlayPoint.id.sid === id.sid);
    }
    refresh(slicesOnly = false) {
        const txt = this.txt;
        let hash = 5381 /* CONST.START_STATE */;
        hash = this.refreshSlices(hash, txt.savedSlices);
        hash = this.refreshSlices(hash, txt.extraSlices);
        hash = this.refreshSlices(hash, txt.localSlices);
        if (!slicesOnly)
            this.computeSplitTextHashes();
        return (this.hash = hash);
    }
    refreshSlices(state, slices) {
        const oldSlicesHash = slices.hash;
        const changed = oldSlicesHash !== slices.refresh();
        const sliceSet = this.slices;
        state = (0, json_hash_1.updateNum)(state, slices.hash);
        if (changed) {
            slices.forEach((slice) => {
                let tuple = sliceSet.get(slice);
                if (tuple) {
                    if (slice.isDel && slice.isDel()) {
                        this.delSlice(slice, tuple);
                        return;
                    }
                    const positionMoved = tuple[0].cmp(slice.start) !== 0 || tuple[1].cmp(slice.end) !== 0;
                    if (positionMoved)
                        this.delSlice(slice, tuple);
                    else
                        return;
                }
                tuple = slice instanceof MarkerSlice_1.MarkerSlice ? this.insMarker(slice) : this.insSlice(slice);
                this.slices.set(slice, tuple);
            });
            if (slices.size() < sliceSet.size) {
                sliceSet.forEach((tuple, slice) => {
                    const mutSlice = slice;
                    if (mutSlice.isDel) {
                        if (!mutSlice.isDel())
                            return;
                        this.delSlice(slice, tuple);
                    }
                });
            }
        }
        return state;
    }
    insSlice(slice) {
        // TODO: Test cases where the inserted slice is collapsed to one point.
        const x0 = slice.start;
        const x1 = slice.end;
        const [start, isStartNew] = this.upsertPoint(x0);
        const [end, isEndNew] = this.upsertPoint(x1);
        const isCollapsed = x0.cmp(x1) === 0;
        start.refs.push(new refs_1.OverlayRefSliceStart(slice));
        end.refs.push(new refs_1.OverlayRefSliceEnd(slice));
        if (isStartNew) {
            const beforeStartPoint = (0, util_1.prev)(start);
            if (beforeStartPoint)
                start.layers.push(...beforeStartPoint.layers);
        }
        if (!isCollapsed) {
            if (isEndNew) {
                const beforeEndPoint = (0, util_1.prev)(end);
                if (beforeEndPoint)
                    end.layers.push(...beforeEndPoint.layers);
            }
            let curr = start;
            do
                curr.addLayer(slice);
            while ((curr = (0, util_1.next)(curr)) && curr !== end);
        }
        else {
            // TODO: review if this is needed:
            start.addMarker(slice);
        }
        return [start, end];
    }
    insMarker(slice) {
        const point = this.mPoint(slice, 0 /* Anchor.Before */);
        const pivot = this.insPoint(point);
        if (!pivot) {
            point.refs.push(slice);
            const prevPoint = (0, util_1.prev)(point);
            if (prevPoint)
                point.layers.push(...prevPoint.layers);
        }
        return [point, point];
    }
    delSlice(slice, [start, end]) {
        this.slices.delete(slice);
        let curr = start;
        do {
            curr.removeLayer(slice);
            curr.removeMarker(slice);
            curr = (0, util_1.next)(curr);
        } while (curr && curr !== end);
        start.removeRef(slice);
        end.removeRef(slice);
        if (!start.refs.length)
            this.delPoint(start);
        if (!end.refs.length && start !== end)
            this.delPoint(end);
    }
    /**
     * Retrieve an existing {@link OverlayPoint} or create a new one, inserted
     * in the tree, sorted by spatial dimension.
     */
    upsertPoint(point) {
        const newPoint = this.point(point.id, point.anchor);
        const pivot = this.insPoint(newPoint);
        if (pivot)
            return [pivot, false];
        return [newPoint, true];
    }
    /**
     * Inserts a point into the tree, sorted by spatial dimension.
     * @param point Point to insert.
     * @returns Returns the existing point if it was already in the tree.
     */
    insPoint(point) {
        let pivot = this.getOrNextLower(point);
        if (!pivot)
            pivot = (0, util_1.first)(this.root);
        if (!pivot) {
            this.root = point;
            return;
        }
        else {
            if (pivot.cmp(point) === 0)
                return pivot;
            const cmp = pivot.cmpSpatial(point);
            if (cmp < 0)
                (0, util_1.insertRight)(point, pivot);
            else
                (0, util_1.insertLeft)(point, pivot);
        }
        if (this.root !== point)
            this.root = (0, util_2.splay)(this.root, point, 10);
        return;
    }
    delPoint(point) {
        this.root = (0, util_1.remove)(this.root, point);
    }
    // ---------------------------------------------------------------- Printable
    toString(tab = '') {
        const printPoint = (tab, point) => {
            return (point.toString(tab) +
                (0, printBinary_1.printBinary)(tab, [
                    !point.l ? null : (tab) => printPoint(tab, point.l),
                    !point.r ? null : (tab) => printPoint(tab, point.r),
                ]));
        };
        return (`${this.constructor.name} #${this.hash.toString(36)}` +
            (0, printTree_1.printTree)(tab, [!this.root ? null : (tab) => printPoint(tab, this.root)]));
    }
}
exports.Overlay = Overlay;
